"""
Module to provide processing for the leaf blocks.
"""
import logging

from pymarkdown.html_helper import HtmlHelper
from pymarkdown.inline_helper import InlineHelper
from pymarkdown.inline_markdown_token import TextMarkdownToken
from pymarkdown.leaf_markdown_token import (
    AtxHeadingMarkdownToken,
    BlankLineMarkdownToken,
    FencedCodeBlockMarkdownToken,
    IndentedCodeBlockMarkdownToken,
    ParagraphMarkdownToken,
    SetextHeadingMarkdownToken,
    ThematicBreakMarkdownToken,
)
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.stack_token import (
    BlockQuoteStackToken,
    FencedCodeBlockStackToken,
    IndentedCodeBlockStackToken,
    ParagraphStackToken,
)

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


class LeafBlockProcessor:
    """
    Class to provide processing for the leaf blocks.
    """

    __fenced_start_tilde = "~"
    __fenced_start_backtick = "`"
    __fenced_code_block_start_characters = (
        f"{__fenced_start_tilde}{__fenced_start_backtick}"
    )
    __thematic_break_characters = "*_-"
    __atx_character = "#"
    __setext_characters = "-="

    @staticmethod
    def is_fenced_code_block(line_to_parse, start_index, extracted_whitespace):
        """
        Determine if we have the start of a fenced code block.
        """

        if (
            ParserHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and ParserHelper.is_character_at_index_one_of(
            line_to_parse,
            start_index,
            LeafBlockProcessor.__fenced_code_block_start_characters,
        ):
            POGGER.debug("ifcb:collected_count>>$<<$<<", line_to_parse, start_index)
            collected_count, new_index = ParserHelper.collect_while_character(
                line_to_parse, start_index, line_to_parse[start_index]
            )
            POGGER.debug("ifcb:collected_count:$", collected_count)
            (
                non_whitespace_index,
                extracted_whitespace_before_info_string,
            ) = ParserHelper.extract_whitespace(line_to_parse, new_index)

            if collected_count >= 3:
                POGGER.debug("ifcb:True")
                return (
                    True,
                    non_whitespace_index,
                    extracted_whitespace_before_info_string,
                    collected_count,
                )
        return False, None, None, None

    # pylint: disable=too-many-locals
    @staticmethod
    def parse_fenced_code_block(
        parser_state,
        position_marker,
        extracted_whitespace,
    ):
        """
        Handle the parsing of a fenced code block
        """

        POGGER.debug(
            "line>>$>>index>>$>>",
            position_marker.text_to_parse,
            position_marker.index_number,
        )
        new_tokens = []
        (
            is_fence_start,
            non_whitespace_index,
            extracted_whitespace_before_info_string,
            collected_count,
        ) = LeafBlockProcessor.is_fenced_code_block(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
        )
        if is_fence_start and not parser_state.token_stack[-1].is_html_block:
            if parser_state.token_stack[-1].is_fenced_code_block:
                POGGER.debug("pfcb->end")

                if (
                    parser_state.token_stack[-1].code_fence_character
                    == position_marker.text_to_parse[position_marker.index_number]
                    and collected_count
                    >= parser_state.token_stack[-1].fence_character_count
                    and non_whitespace_index >= len(position_marker.text_to_parse)
                ):
                    new_end_token = parser_state.token_stack[
                        -1
                    ].generate_close_markdown_token_from_stack_token(
                        extracted_whitespace, extra_end_data=str(collected_count)
                    )
                    new_tokens.append(new_end_token)
                    del parser_state.token_stack[-1]
            else:
                POGGER.debug("pfcb->check")
                if (
                    position_marker.text_to_parse[position_marker.index_number]
                    == LeafBlockProcessor.__fenced_start_tilde
                    or LeafBlockProcessor.__fenced_start_backtick
                    not in position_marker.text_to_parse[non_whitespace_index:]
                ):
                    POGGER.debug("pfcb->start")
                    (
                        after_extracted_text_index,
                        extracted_text,
                    ) = ParserHelper.extract_until_whitespace(
                        position_marker.text_to_parse, non_whitespace_index
                    )
                    text_after_extracted_text = position_marker.text_to_parse[
                        after_extracted_text_index:
                    ]

                    old_top_of_stack = parser_state.token_stack[-1]
                    new_tokens, _ = parser_state.close_open_blocks_fn(
                        parser_state,
                        only_these_blocks=[ParagraphStackToken],
                    )

                    pre_extracted_text, pre_text_after_extracted_text = (
                        extracted_text,
                        text_after_extracted_text,
                    )

                    extracted_text = InlineHelper.handle_backslashes(
                        extracted_text, add_text_signature=False
                    )
                    text_after_extracted_text = InlineHelper.handle_backslashes(
                        text_after_extracted_text, add_text_signature=False
                    )

                    if pre_extracted_text == extracted_text:
                        pre_extracted_text = ""
                    if pre_text_after_extracted_text == text_after_extracted_text:
                        pre_text_after_extracted_text = ""

                    new_token = FencedCodeBlockMarkdownToken(
                        position_marker.text_to_parse[position_marker.index_number],
                        collected_count,
                        extracted_text,
                        pre_extracted_text,
                        text_after_extracted_text,
                        pre_text_after_extracted_text,
                        extracted_whitespace,
                        extracted_whitespace_before_info_string,
                        position_marker,
                    )
                    new_tokens.append(new_token)
                    parser_state.token_stack.append(
                        FencedCodeBlockStackToken(
                            code_fence_character=position_marker.text_to_parse[
                                position_marker.index_number
                            ],
                            fence_character_count=collected_count,
                            whitespace_start_count=ParserHelper.calculate_length(
                                extracted_whitespace
                            ),
                            matching_markdown_token=new_token,
                        )
                    )
                    POGGER.debug("StackToken-->$<<", parser_state.token_stack[-1])
                    POGGER.debug(
                        "StackToken>start_markdown_token-->$<<",
                        parser_state.token_stack[-1].matching_markdown_token,
                    )

                    LeafBlockProcessor.correct_for_leaf_block_start_in_list(
                        parser_state,
                        position_marker.index_indent,
                        old_top_of_stack,
                        new_tokens,
                    )

        elif (
            parser_state.token_stack[-1].is_fenced_code_block
            and parser_state.token_stack[-1].whitespace_start_count
            and extracted_whitespace
        ):

            current_whitespace_length = ParserHelper.calculate_length(
                extracted_whitespace
            )
            whitespace_left = max(
                0,
                current_whitespace_length
                - parser_state.token_stack[-1].whitespace_start_count,
            )
            POGGER.debug("previous_ws>>$", current_whitespace_length)
            POGGER.debug("whitespace_left>>$", whitespace_left)
            removed_whitespace = ParserHelper.create_replace_with_nothing_marker(
                ParserHelper.repeat_string(
                    ParserHelper.space_character,
                    current_whitespace_length - whitespace_left,
                )
            )
            extracted_whitespace = f"{removed_whitespace}{ParserHelper.repeat_string(ParserHelper.space_character, whitespace_left)}"
        return new_tokens, extracted_whitespace

    # pylint: enable=too-many-locals

    @staticmethod
    def __adjust_for_list_start(
        parser_state,
        last_list_start_index,
        last_block_quote_index,
    ):
        did_process, offset_index = False, 0
        POGGER.debug("last_list_start_index>>$>>", last_list_start_index)
        POGGER.debug(
            "parser_state.original_line_to_parse>>$>>",
            parser_state.original_line_to_parse,
        )
        if last_list_start_index:
            new_index, extracted_whitespace = ParserHelper.extract_whitespace_from_end(
                parser_state.original_line_to_parse, last_list_start_index
            )
            POGGER.debug("new_index>>$>>", new_index)
            POGGER.debug("extracted_whitespace>>$>>", extracted_whitespace)
            if ParserHelper.tab_character in extracted_whitespace:
                did_process, last_block_quote_index, offset_index = True, new_index, 1
        return did_process, offset_index, last_block_quote_index

    # pylint: disable=too-many-locals
    @staticmethod
    def __adjust_for_block_quote_start(
        force_me,
        parser_state,
        last_block_quote_index,
        position_marker,
        extracted_whitespace,
    ):
        """
        Block quotes cause indents, which need to be handled specifically.
        """

        (
            did_process,
            special_parse_start_index,
            whitespace_to_parse,
            block_quote_adjust_delta,
        ) = (False, 0, extracted_whitespace, 0)
        POGGER.debug(
            "last_block_quote_index>>$>>force_me>>$",
            last_block_quote_index,
            force_me,
        )
        if last_block_quote_index or force_me:
            POGGER.debug(
                "original_line_to_parse>[$]>>last_block_quote_index>>$",
                parser_state.original_line_to_parse,
                last_block_quote_index,
            )
            (
                block_quote_after_whitespace_index,
                during_original_whitespace,
            ) = ParserHelper.extract_whitespace(
                parser_state.original_line_to_parse, last_block_quote_index
            )
            POGGER.debug(
                "during_original_whitespace>[$]",
                during_original_whitespace,
            )
            if ParserHelper.tab_character in during_original_whitespace:

                did_process = True
                POGGER.debug(
                    ".text_to_parse>[$]",
                    position_marker,
                )
                POGGER.debug(".index_number>>$", position_marker.index_number)
                POGGER.debug(".index_indent>>$", position_marker.index_indent)
                POGGER.debug("last_block_quote_index>>$", last_block_quote_index)

                # Make sure everything after the whitespace remains the same.
                text_after_original_whitespace, text_after_whitespace = (
                    parser_state.original_line_to_parse[
                        block_quote_after_whitespace_index:
                    ],
                    position_marker.text_to_parse[position_marker.index_number :],
                )
                POGGER.debug(
                    "text_after_original_whitespace>[$]", text_after_original_whitespace
                )
                POGGER.debug(
                    "text_after_whitespace>[$]",
                    text_after_whitespace,
                )
                assert text_after_original_whitespace == text_after_whitespace

                # Make sure the whitespace is within expected bounds.
                during_current_whitespace = position_marker.text_to_parse[
                    position_marker.index_number
                    - len(extracted_whitespace) : position_marker.index_number
                ]
                POGGER.debug(
                    "during_current_whitespace>[$]",
                    during_current_whitespace,
                )
                POGGER.debug(
                    "during_original_whitespace>[$]",
                    during_original_whitespace,
                )

                current_whitespace_length, original_whitespace_length = len(
                    during_current_whitespace
                ), (
                    ParserHelper.calculate_length(
                        during_original_whitespace, start_index=last_block_quote_index
                    )
                    - 1
                )
                POGGER.debug(
                    "current_whitespace_length[$],original_whitespace_length[$]",
                    current_whitespace_length,
                    original_whitespace_length,
                )
                assert current_whitespace_length <= original_whitespace_length

                special_parse_start_index = last_block_quote_index + 1
                if during_original_whitespace[0] == ParserHelper.tab_character:
                    whitespace_to_parse = during_original_whitespace
                    if (
                        current_whitespace_length > 1
                        and during_original_whitespace[1] == ParserHelper.tab_character
                    ):
                        block_quote_adjust_delta = -1
                else:
                    whitespace_to_parse = during_original_whitespace[1:]

        return (
            did_process,
            special_parse_start_index,
            whitespace_to_parse,
            block_quote_adjust_delta,
        )

    # pylint: enable=too-many-locals

    @staticmethod
    def __recalculate_whitespace(
        special_parse_start_index, whitespace_to_parse, offset_index
    ):
        """
        Recalculate the whitespace characteristics.
        """

        accumulated_whitespace_count = 0
        actual_whitespace_index = 0
        abc = 4 + offset_index

        relative_whitespace_index = special_parse_start_index - offset_index
        POGGER.debug(
            "whitespace_to_parse>>$>>",
            whitespace_to_parse,
        )
        POGGER.debug("special_parse_start_index>>$>>", special_parse_start_index)
        POGGER.debug(
            "in>>index>>$($)>>accumulated_whitespace_count>>$",
            actual_whitespace_index,
            actual_whitespace_index + special_parse_start_index,
            accumulated_whitespace_count,
        )
        while accumulated_whitespace_count < abc:
            if (
                whitespace_to_parse[actual_whitespace_index]
                == ParserHelper.tab_character
            ):
                POGGER.debug(
                    ">>relative_whitespace_index>>$", relative_whitespace_index
                )
                delta_whitespace = 4 - (relative_whitespace_index % 4)
            else:
                delta_whitespace = 1
            POGGER.debug(">>delta_whitespace>>$", delta_whitespace)
            accumulated_whitespace_count += delta_whitespace
            relative_whitespace_index += delta_whitespace
            actual_whitespace_index += 1
            POGGER.debug(
                ">>index>>$($)>>accumulated_whitespace_count>>$",
                actual_whitespace_index,
                (actual_whitespace_index + special_parse_start_index),
                accumulated_whitespace_count,
            )

        POGGER.debug(
            "out>>index>>$($)>>accumulated_whitespace_count>>$",
            actual_whitespace_index,
            (actual_whitespace_index + special_parse_start_index),
            accumulated_whitespace_count,
        )

        adj_ws = whitespace_to_parse[0:actual_whitespace_index]
        left_ws = whitespace_to_parse[actual_whitespace_index:]
        POGGER.debug("accumulated_whitespace_count>>$", accumulated_whitespace_count)
        POGGER.debug("actual_whitespace_index>>$", actual_whitespace_index)
        POGGER.debug("adj_ws>>$<<", adj_ws)
        POGGER.debug("left_ws>>$<<", left_ws)
        POGGER.debug("offset_index>>$<<", offset_index)

        return accumulated_whitespace_count, actual_whitespace_index, adj_ws, left_ws

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def parse_indented_code_block(
        parser_state,
        position_marker,
        extracted_whitespace,
        removed_chars_at_start,
        last_block_quote_index,
        last_list_start_index,
    ):
        """
        Handle the parsing of an indented code block
        """

        new_tokens = []

        if (
            ParserHelper.is_length_greater_than_or_equal_to(
                extracted_whitespace, 4, start_index=removed_chars_at_start
            )
            and not parser_state.token_stack[-1].is_paragraph
        ):
            if not parser_state.token_stack[-1].is_indented_code_block:
                POGGER.debug(">>__adjust_for_list_start")
                (
                    did_process,
                    offset_index,
                    last_block_quote_index,
                ) = LeafBlockProcessor.__adjust_for_list_start(
                    parser_state,
                    last_list_start_index,
                    last_block_quote_index,
                )
                POGGER.debug("<<__adjust_for_list_start<<$", did_process)

                force_me = False
                kludge_adjust = 0
                if not did_process:
                    POGGER.debug(">>>>$", parser_state.token_stack[-1])
                    if parser_state.token_stack[-1].is_list:
                        POGGER.debug(
                            ">>indent>>$",
                            parser_state.token_stack[-1].indent_level,
                        )
                        last_block_quote_index = 0
                        kludge_adjust = 1
                        force_me = True

                POGGER.debug(">>__adjust_for_block_quote_start")
                (
                    did_process,
                    special_parse_start_index,
                    whitespace_to_parse,
                    block_quote_adjust_delta,
                ) = LeafBlockProcessor.__adjust_for_block_quote_start(
                    force_me,
                    parser_state,
                    last_block_quote_index,
                    position_marker,
                    extracted_whitespace,
                )
                POGGER.debug("<<__adjust_for_block_quote_start<<$", did_process)

                POGGER.debug(
                    "__recalculate_whitespace>>$>>$",
                    whitespace_to_parse,
                    offset_index,
                )
                (
                    accumulated_whitespace_count,
                    actual_whitespace_index,
                    adj_ws,
                    left_ws,
                ) = LeafBlockProcessor.__recalculate_whitespace(
                    special_parse_start_index, whitespace_to_parse, offset_index
                )

                # TODO revisit with tabs
                line_number = position_marker.line_number
                column_number = (
                    position_marker.index_number
                    + position_marker.index_indent
                    - len(extracted_whitespace)
                    + 1
                )
                if special_parse_start_index:
                    column_number = (
                        actual_whitespace_index
                        + special_parse_start_index
                        + block_quote_adjust_delta
                    )
                    POGGER.debug(
                        "column_number($)=actual_whitespace_index($)+special_parse_start_index($)+block_quote_adjust_delta($)",
                        column_number,
                        actual_whitespace_index,
                        special_parse_start_index,
                        block_quote_adjust_delta,
                    )
                    excess_whitespace_count = (
                        accumulated_whitespace_count - 4 - offset_index
                    )
                    POGGER.debug(
                        "excess_whitespace_count($)=accumulated_whitespace_count($)-4-offset_index($)",
                        excess_whitespace_count,
                        accumulated_whitespace_count,
                        offset_index,
                    )
                    POGGER.debug("before>>$>>", left_ws)
                    if excess_whitespace_count:
                        excess_whitespace_count -= kludge_adjust
                        left_ws = f"{ParserHelper.repeat_string(ParserHelper.space_character, excess_whitespace_count)}{left_ws}"
                    POGGER.debug("after>>$>>", left_ws)
                else:
                    column_number += actual_whitespace_index
                POGGER.debug("column_number>>$", column_number)

                new_token = IndentedCodeBlockMarkdownToken(
                    adj_ws, line_number, column_number
                )
                parser_state.token_stack.append(IndentedCodeBlockStackToken(new_token))
                new_tokens.append(new_token)
                extracted_whitespace = left_ws
                POGGER.debug("left_ws>>$<<", extracted_whitespace)
            new_tokens.append(
                TextMarkdownToken(
                    position_marker.text_to_parse[position_marker.index_number :],
                    extracted_whitespace,
                    position_marker=position_marker,
                )
            )
        return new_tokens

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def is_thematic_break(
        line_to_parse,
        start_index,
        extracted_whitespace,
        skip_whitespace_check=False,
        whitespace_allowed_between_characters=True,
    ):
        """
        Determine whether or not we have a thematic break.
        """

        thematic_break_character, end_of_break_index = None, None
        is_thematic_character = ParserHelper.is_character_at_index_one_of(
            line_to_parse, start_index, LeafBlockProcessor.__thematic_break_characters
        )
        POGGER.debug("skip_whitespace_check>>$", skip_whitespace_check)
        POGGER.debug("is_thematic_character>>$", is_thematic_character)
        if (
            ParserHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
            or skip_whitespace_check
        ) and is_thematic_character:
            start_char, index, char_count, repeat_loop, line_to_parse_size = (
                line_to_parse[start_index],
                start_index,
                0,
                True,
                len(line_to_parse),
            )
            while repeat_loop and index < line_to_parse_size:
                if (
                    whitespace_allowed_between_characters
                    and ParserHelper.is_character_at_index_whitespace(
                        line_to_parse, index
                    )
                ):
                    index += 1
                elif line_to_parse[index] == start_char:
                    index += 1
                    char_count += 1
                else:
                    repeat_loop = False

            POGGER.debug("char_count>>$", char_count)
            POGGER.debug("index>>$", index)
            POGGER.debug("line_to_parse_size>>$", line_to_parse_size)
            if char_count >= 3 and index == line_to_parse_size:
                thematic_break_character, end_of_break_index = start_char, index

        return thematic_break_character, end_of_break_index

    @staticmethod
    def parse_thematic_break(
        parser_state,
        position_marker,
        extracted_whitespace,
        this_bq_count,
        stack_bq_count,
    ):
        """
        Handle the parsing of a thematic break.
        """

        new_tokens = []

        start_char, index = LeafBlockProcessor.is_thematic_break(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
        )
        if start_char:
            # TODO why not use close?
            force_paragraph_close_if_present = this_bq_count == 0 and stack_bq_count > 0
            if parser_state.token_stack[-1].is_paragraph:
                new_tokens.append(
                    parser_state.token_stack[
                        -1
                    ].generate_close_markdown_token_from_stack_token(
                        was_forced=force_paragraph_close_if_present
                    )
                )
                del parser_state.token_stack[-1]
            if this_bq_count == 0 and stack_bq_count > 0:
                new_tokens, _ = parser_state.close_open_blocks_fn(
                    parser_state,
                    destination_array=new_tokens,
                    only_these_blocks=[BlockQuoteStackToken],
                    include_block_quotes=True,
                    was_forced=True,
                )
            new_tokens.append(
                ThematicBreakMarkdownToken(
                    start_char,
                    extracted_whitespace.replace(ParserHelper.tab_character, "    "),
                    position_marker.text_to_parse[
                        position_marker.index_number : index
                    ].replace(ParserHelper.tab_character, "    "),
                    position_marker=position_marker,
                )
            )
        return new_tokens

    @staticmethod
    def is_atx_heading(line_to_parse, start_index, extracted_whitespace):
        """
        Determine whether or not an ATX Heading is about to start.
        """

        if ParserHelper.is_length_less_than_or_equal_to(
            extracted_whitespace, 3
        ) and ParserHelper.is_character_at_index(
            line_to_parse,
            start_index,
            LeafBlockProcessor.__atx_character,
        ):
            hash_count, new_index = ParserHelper.collect_while_character(
                line_to_parse,
                start_index,
                LeafBlockProcessor.__atx_character,
            )

            non_whitespace_index = ParserHelper.collect_while_character(
                line_to_parse, new_index, " "
            )
            extracted_whitespace_at_start = line_to_parse[
                new_index : non_whitespace_index[1]
            ]

            if hash_count <= 6 and (
                extracted_whitespace_at_start
                or non_whitespace_index[1] == len(line_to_parse)
            ):
                return (
                    True,
                    non_whitespace_index[1],
                    hash_count,
                    extracted_whitespace_at_start,
                )
        return False, None, None, None

    # pylint: disable=too-many-locals
    @staticmethod
    def parse_atx_headings(parser_state, position_marker, extracted_whitespace):
        """
        Handle the parsing of an atx heading.
        """

        new_tokens = []

        (
            heading_found,
            non_whitespace_index,
            hash_count,
            extracted_whitespace_at_start,
        ) = LeafBlockProcessor.is_atx_heading(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
        )
        if heading_found:
            POGGER.debug(
                "parse_atx_headings>>start",
            )

            (
                old_top_of_stack,
                remaining_line,
                remove_trailing_count,
                extracted_whitespace_before_end,
            ) = (
                parser_state.token_stack[-1],
                position_marker.text_to_parse[non_whitespace_index:],
                0,
                "",
            )

            new_tokens, _ = parser_state.close_open_blocks_fn(parser_state, new_tokens)
            (
                end_index,
                extracted_whitespace_at_end,
            ) = ParserHelper.extract_whitespace_from_end(remaining_line)
            while (
                end_index > 0
                and remaining_line[end_index - 1] == LeafBlockProcessor.__atx_character
            ):
                end_index -= 1
                remove_trailing_count += 1
            if remove_trailing_count:
                if end_index > 0:
                    if ParserHelper.is_character_at_index(
                        remaining_line, end_index - 1, " "
                    ):
                        remaining_line = remaining_line[:end_index]
                        non_whitespace_index = (
                            ParserHelper.collect_backwards_while_character(
                                remaining_line, len(remaining_line) - 1, " "
                            )
                        )
                        end_index = non_whitespace_index[1]
                        extracted_whitespace_before_end = remaining_line[end_index:]
                        remaining_line = remaining_line[:end_index]
                    else:
                        extracted_whitespace_at_end, remove_trailing_count = "", 0
                else:
                    remaining_line = ""
            else:
                extracted_whitespace_at_end = remaining_line[end_index:]
                remaining_line = remaining_line[0:end_index]
            start_token = AtxHeadingMarkdownToken(
                hash_count,
                remove_trailing_count,
                extracted_whitespace,
                position_marker,
            )
            new_tokens.append(start_token)

            LeafBlockProcessor.correct_for_leaf_block_start_in_list(
                parser_state,
                position_marker.index_indent,
                old_top_of_stack,
                new_tokens,
                was_token_already_added_to_stack=False,
            )

            new_tokens.append(
                TextMarkdownToken(
                    remaining_line,
                    extracted_whitespace_at_start,
                    position_marker=position_marker,
                )
            )
            end_token = start_token.generate_close_markdown_token_from_markdown_token(
                extracted_whitespace_at_end, extracted_whitespace_before_end, False
            )
            new_tokens.append(end_token)
        return new_tokens

    # pylint: enable=too-many-locals

    @staticmethod
    def parse_setext_headings(
        parser_state,
        position_marker,
        extracted_whitespace,
        this_bq_count,
        stack_bq_count,
    ):
        """
        Handle the parsing of an setext heading.
        """

        new_tokens = []
        if (
            ParserHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
            and ParserHelper.is_character_at_index_one_of(
                position_marker.text_to_parse,
                position_marker.index_number,
                LeafBlockProcessor.__setext_characters,
            )
            and parser_state.token_stack[-1].is_paragraph
            and (this_bq_count == stack_bq_count)
        ):
            _, collected_to_index = ParserHelper.collect_while_character(
                position_marker.text_to_parse,
                position_marker.index_number,
                position_marker.text_to_parse[position_marker.index_number],
            )
            (
                after_whitespace_index,
                extra_whitespace_after_setext,
            ) = ParserHelper.extract_whitespace(
                position_marker.text_to_parse, collected_to_index
            )
            if after_whitespace_index == len(position_marker.text_to_parse):

                token_index = len(parser_state.token_document) - 1
                while not parser_state.token_document[token_index].is_paragraph:
                    token_index -= 1

                replacement_token = SetextHeadingMarkdownToken(
                    position_marker.text_to_parse[position_marker.index_number],
                    collected_to_index - position_marker.index_number,
                    parser_state.token_document[token_index].extra_data,
                    position_marker,
                    parser_state.token_document[token_index],
                )

                # This is unusual.  Normally, close_open_blocks is used to close off
                # blocks based on the stack token.  However, since the setext takes
                # the last paragraph of text (see case 61) and translates it
                # into a heading, this has to be done separately, as there is no
                # stack token to close.
                new_tokens.append(
                    replacement_token.generate_close_markdown_token_from_markdown_token(
                        extracted_whitespace, extra_whitespace_after_setext, False
                    )
                )

                parser_state.token_document[token_index] = replacement_token
                del parser_state.token_stack[-1]
        return new_tokens

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    @staticmethod
    def parse_paragraph(
        parser_state,
        position_marker,
        extracted_whitespace,
        this_bq_count,
        stack_bq_count,
        text_removed_by_container,
        force_it,
    ):
        """
        Handle the parsing of a paragraph.
        """
        new_tokens = []

        if parser_state.no_para_start_if_empty and position_marker.index_number >= len(
            position_marker.text_to_parse
        ):
            POGGER.debug("Escaping paragraph due to empty w/ blank")
            return [
                BlankLineMarkdownToken(
                    extracted_whitespace, position_marker, len(extracted_whitespace)
                )
            ]

        POGGER.debug(
            "parse_paragraph>stack_bq_count>$>this_bq_count>$<",
            stack_bq_count,
            this_bq_count,
        )

        container_index, adjusted_whitespace_length = (
            parser_state.find_last_container_on_stack(),
            0,
        )
        if container_index > 0:
            if parser_state.token_stack[container_index].is_block_quote:
                top_block_token = parser_state.token_stack[container_index]
                POGGER.debug(">>block-owners>>$", top_block_token)
                LeafBlockProcessor.__adjust_paragraph_for_block_quotes(
                    top_block_token,
                    text_removed_by_container,
                    force_it,
                    parser_state.token_document,
                )
            else:
                top_list_token = parser_state.token_stack[container_index]
                POGGER.debug(">>list-owners>>$", top_list_token)
                adjusted_whitespace_length = (
                    LeafBlockProcessor.__adjust_paragraph_for_list(
                        top_list_token, extracted_whitespace
                    )
                )

        # In cases where the list ended on the same line as we are processing, the
        # container tokens will not yet be added to the token_document.  As such,
        # make sure to construct a "proper" list that takes those into account
        # before checking to see if this is an issue.
        adjusted_document = parser_state.token_document[:]
        if parser_state.same_line_container_tokens:
            adjusted_document.extend(parser_state.same_line_container_tokens)

        if (
            len(adjusted_document) >= 2
            and adjusted_document[-1].is_blank_line
            and adjusted_document[-2].is_any_list_token
        ):

            did_find, last_list_index = LeafBlockProcessor.check_for_list_in_process(
                parser_state
            )
            assert did_find
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state, until_this_index=last_list_index
            )

        if stack_bq_count != 0 and this_bq_count == 0:
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[BlockQuoteStackToken],
                include_block_quotes=True,
            )

        if adjusted_whitespace_length:
            POGGER.debug(">>GGHHJJ!!>>$>>", adjusted_whitespace_length)
            extracted_whitespace = ParserHelper.repeat_string(
                ParserHelper.blech_character, adjusted_whitespace_length
            )

        if not parser_state.token_stack[-1].is_paragraph:
            new_paragraph_token = ParagraphMarkdownToken(
                extracted_whitespace, position_marker
            )
            parser_state.token_stack.append(ParagraphStackToken(new_paragraph_token))
            new_tokens.append(new_paragraph_token)
            extracted_whitespace = ""

        new_tokens.append(
            TextMarkdownToken(
                position_marker.text_to_parse[position_marker.index_number :],
                extracted_whitespace,
                position_marker=position_marker,
            )
        )
        return new_tokens

    # pylint: enable=too-many-arguments
    # pylint: enable=too-many-locals

    @staticmethod
    def __adjust_paragraph_for_list(top_list_token, extracted_whitespace):
        ex_ws_length = len(extracted_whitespace)
        POGGER.debug(">>owners-indent>>$", top_list_token.indent_level)
        POGGER.debug(">>ws_before_marker>>$", top_list_token.ws_before_marker)
        POGGER.debug(">>ws_after_marker>>$", top_list_token.ws_after_marker)
        POGGER.debug(">>last_new_list_token>>$", top_list_token.last_new_list_token)
        POGGER.debug(">>extracted_whitespace>>$", ex_ws_length)

        dominant_indent, original_list_indent, indent_delta = (
            (
                top_list_token.last_new_list_token.indent_level
                if top_list_token.last_new_list_token
                else top_list_token.indent_level
            ),
            top_list_token.indent_level - 2,
            (
                top_list_token.ws_after_marker - 1
                if top_list_token.ws_after_marker > 1
                else 0
            ),
        )
        POGGER.debug(">>dominant_indent>>$>>", dominant_indent)
        original_text_indent = (
            ex_ws_length
            + top_list_token.indent_level
            - top_list_token.ws_before_marker
            - indent_delta
        )
        POGGER.debug(">>original_list_indent>>$>>", original_list_indent)
        POGGER.debug(">>original_text_indent>$>>", original_text_indent)
        adjusted_whitespace_length = (
            dominant_indent - original_text_indent
            if dominant_indent > original_text_indent >= 4
            else 0
        )
        return adjusted_whitespace_length

    @staticmethod
    def __adjust_paragraph_for_block_quotes(
        top_block_token,
        text_removed_by_container,
        force_it,
        token_document,
    ):
        POGGER.debug(">>list-owners>>$", token_document)
        number_of_block_quote_ends, end_index = 0, len(token_document) - 1
        while token_document[end_index].is_block_quote_end:
            number_of_block_quote_ends += 1
            end_index -= 1
        if not (
            number_of_block_quote_ends > 0
            and token_document[end_index].is_fenced_code_block_end
        ):
            if text_removed_by_container is None:
                top_block_token.matching_markdown_token.add_leading_spaces("")
            elif force_it:
                top_block_token.matching_markdown_token.add_leading_spaces(
                    text_removed_by_container
                )

    @staticmethod
    def check_for_list_in_process(parser_state):
        """
        From the end of the stack, check to see if there is already a list in progress.
        """

        stack_index = len(parser_state.token_stack) - 1

        while stack_index >= 0 and not parser_state.token_stack[stack_index].is_list:
            stack_index -= 1

        return stack_index >= 0, stack_index

    @staticmethod
    def correct_for_leaf_block_start_in_list(
        parser_state,
        removed_chars_at_start,
        old_top_of_stack,
        html_tokens,
        was_token_already_added_to_stack=True,
    ):
        """
        Check to see that if a paragraph has been closed within a list and
        there is a leaf block token immediately following, that the right
        actions are taken.
        """

        POGGER.debug(">>__xx>>removed_chars_at_start>$>>", removed_chars_at_start)
        if not old_top_of_stack.is_paragraph:
            POGGER.debug("1")
            return html_tokens

        statck_index, top_of_stack, end_of_list = (
            -2 if was_token_already_added_to_stack else -1,
            None,
            html_tokens[-1],
        )
        if not parser_state.token_stack[statck_index].is_list:
            POGGER.debug("2")
            return html_tokens

        POGGER.debug(">>__xx>>stack>>$>>", parser_state.token_stack)
        POGGER.debug(">>__xx>>tokens>>$>>", parser_state.token_document)
        POGGER.debug(">>__xx>>tokens_to_add>>$>>", html_tokens)

        if was_token_already_added_to_stack:
            top_of_stack = parser_state.token_stack[-1]
            del parser_state.token_stack[-1]
        del html_tokens[-1]

        POGGER.debug(">>__xx>>stack>>$>>", parser_state.token_stack)
        POGGER.debug(">>__xx>>tokens_to_add>>$>>", html_tokens)

        repeat_loop = True
        is_remaining_list_token = True
        while repeat_loop and is_remaining_list_token:
            assert parser_state.token_stack[-1].is_list

            POGGER.debug(">>removed_chars_at_start>>$>>", removed_chars_at_start)
            POGGER.debug(
                ">>stack indent>>$>>", parser_state.token_stack[-1].indent_level
            )
            if removed_chars_at_start >= parser_state.token_stack[-1].indent_level:
                repeat_loop = False
            else:
                tokens_from_close, _ = parser_state.close_open_blocks_fn(
                    parser_state,
                    until_this_index=(len(parser_state.token_stack) - 1),
                    include_lists=True,
                )
                POGGER.debug(">>__xx>>tokens_from_close>>$>>", tokens_from_close)
                html_tokens.extend(tokens_from_close)

                is_remaining_list_token = parser_state.token_stack[-1].is_list

        if is_remaining_list_token:
            assert parser_state.token_stack[-1].is_list
            last_indent = parser_state.token_stack[-1].indent_level
            delta_indent = removed_chars_at_start - last_indent
            POGGER.debug(">>__xx>>delta_indent>>$>>", delta_indent)
            assert not delta_indent

        if was_token_already_added_to_stack:
            parser_state.token_stack.append(top_of_stack)
            POGGER.debug(">>__xx>>stack>>$>>", parser_state.token_stack)
        html_tokens.append(end_of_list)
        POGGER.debug(">>__xx>>tokens_to_add>>$>>", html_tokens)

    @staticmethod
    def is_paragraph_ending_leaf_block_start(
        parser_state,
        line_to_parse,
        start_index,
        extracted_whitespace,
        exclude_thematic_break=False,
    ):
        """
        Determine whether we have a valid leaf block start.
        """

        is_leaf_block_start = False
        if not exclude_thematic_break:
            is_leaf_block_start, _ = LeafBlockProcessor.is_thematic_break(
                line_to_parse,
                start_index,
                extracted_whitespace,
                skip_whitespace_check=True,
            )
            is_leaf_block_start = bool(is_leaf_block_start)
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_theme_break>>$",
                is_leaf_block_start,
            )
        if not is_leaf_block_start:
            is_leaf_block_start, _ = HtmlHelper.is_html_block(
                line_to_parse,
                start_index,
                extracted_whitespace,
                parser_state.token_stack,
            )
            is_leaf_block_start = bool(is_leaf_block_start)
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_html_block>>$",
                is_leaf_block_start,
            )
        if not is_leaf_block_start:
            is_leaf_block_start, _, _, _ = LeafBlockProcessor.is_fenced_code_block(
                line_to_parse, start_index, extracted_whitespace
            )
            is_leaf_block_start = bool(is_leaf_block_start)
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_fenced_code_block>>$",
                is_leaf_block_start,
            )
        if not is_leaf_block_start:
            is_leaf_block_start, _, _, _ = LeafBlockProcessor.is_atx_heading(
                line_to_parse, start_index, extracted_whitespace
            )
            is_leaf_block_start = bool(is_leaf_block_start)
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_atx_heading>>$",
                is_leaf_block_start,
            )
        return is_leaf_block_start



"""
Module to implement a plugin that looks for more than one space between either the
opening or closing hashes of an atx heading.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd021(Plugin):
    """
    Class to implement a plugin that looks for more than one space between either the
    opening or closing hashes of an atx heading.
    """

    def __init__(self):
        super().__init__()
        self.__atx_heading_token = None
        self.__is_left_in_error = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-multiple-space-closed-atx",
            plugin_id="MD021",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces are present inside hash characters on Atx Closed Heading.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md021.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__atx_heading_token = None
        self.__is_left_in_error = False

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading:
            if token.remove_trailing_count:
                self.__atx_heading_token = token
            self.__is_left_in_error = False
        elif token.is_paragraph_end:
            self.__atx_heading_token = None
        elif token.is_atx_heading_end:
            if self.__is_left_in_error or len(token.extra_end_data) > 1:
                self.report_next_token_error(context, self.__atx_heading_token)
        elif token.is_text:
            if self.__atx_heading_token and len(token.extracted_whitespace) > 1:
                self.__is_left_in_error = True



"""
Module to provide for a leaf element that can be added to markdown parsing stream that handles front matter.
"""
import logging
import string

from pymarkdown.extension_impl import ExtensionDetails
from pymarkdown.leaf_block_processor import LeafBlockProcessor
from pymarkdown.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.markdown_token import MarkdownToken
from pymarkdown.parser_helper import ParserHelper, PositionMarker
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


class FrontMatterExtension:
    """
    Extension to implement the front matter extensions.
    """

    @classmethod
    def get_identifier(cls):
        """
        Get the identifier associated with this extension.
        """
        return "front-matter"

    @classmethod
    def get_details(cls):
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=cls.get_identifier(),
            extension_name="Front Matter Metadata",
            extension_description="Allows metadata to be parsed from document front matter.",
            extension_enabled_by_default=False,
            extension_version="0.5.0",
            extension_interface_version=1,
            extension_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/extensions/front-matter.md",
            extension_configuration=None,
        )

    @classmethod
    def apply_configuration(cls, extension_specific_facade):
        """
        Apply any configuration required by the extension.
        """
        _ = extension_specific_facade

    @staticmethod
    def handle_front_matter_token(output_html, next_token, transform_state):
        """
        Handle the front matter token.  Note that it does not contribute anything
        at all to the HTML output.
        """
        _ = (next_token, transform_state)

        return output_html

    @staticmethod
    def rehydrate_front_matter(current_token, previous_token):
        """
        Rehydrate the front matter text from the token.
        """
        _ = previous_token

        front_matter_parts = [current_token.start_boundary_line]
        front_matter_parts.extend(current_token.collected_lines)
        front_matter_parts.extend([current_token.end_boundary_line, ""])
        return ParserHelper.newline_character.join(front_matter_parts)

    @staticmethod
    def process_header_if_present(
        token_to_use, line_number, requeue, source_provider, tokenized_document
    ):
        """
        Take care of processing eligibility and processing for front matter support.
        """

        start_char, extracted_index = LeafBlockProcessor.is_thematic_break(
            token_to_use.rstrip(),
            0,
            "",
            whitespace_allowed_between_characters=False,
            skip_whitespace_check=True,
        )
        if start_char == "-" and extracted_index == 3:
            (
                token_to_use,
                new_token,
                line_number,
                requeue_lines,
            ) = FrontMatterExtension.__handle_document_front_matter(
                token_to_use, source_provider
            )
            if new_token:
                tokenized_document.append(new_token)
            else:
                requeue.extend(requeue_lines)
                token_to_use = requeue[0]
                del requeue[0]
            POGGER.debug("self.tokenized_document>>$", tokenized_document)
            POGGER.debug("requeue>>$", requeue)
        return token_to_use, line_number, requeue

    @staticmethod
    def __handle_document_front_matter(token_to_use, source_provider):

        starting_line = token_to_use
        clean_starting_line = starting_line.rstrip()
        repeat_again = True
        have_closing = False
        collected_lines = []
        POGGER.info("Metadata prefix detected, scanning for metadata header.")
        while repeat_again:
            token_to_use = source_provider.get_next_line()
            if token_to_use and token_to_use.rstrip():
                start_char, _ = LeafBlockProcessor.is_thematic_break(
                    token_to_use.rstrip(),
                    0,
                    "",
                    whitespace_allowed_between_characters=False,
                )
                have_closing = (
                    start_char and clean_starting_line == token_to_use.rstrip()
                )
                repeat_again = not have_closing
            else:
                repeat_again = token_to_use is not None
            if repeat_again:
                collected_lines.append(token_to_use)

        if not have_closing:
            POGGER.info(
                "Metadata prefix abandoned. End of document reached before closing fence encountered."
            )
            collected_lines.insert(0, starting_line)
            return None, None, 1, collected_lines

        POGGER.info("Metadata prefix collected. Verifying validity.")
        matter_map = FrontMatterExtension.__is_front_matter_valid(collected_lines)
        POGGER.debug("ret=$s,type=$s", matter_map, type(matter_map))
        if isinstance(matter_map, str):
            POGGER.info("Metadata validation failed: $", matter_map)
            collected_lines.insert(0, starting_line)
            collected_lines.append(starting_line)
            return None, None, 1, collected_lines

        POGGER.info("Metadata validation succeeded.")
        position_marker = PositionMarker(1, 0, starting_line)
        new_token = FrontMatterMarkdownToken(
            starting_line, token_to_use, collected_lines, matter_map, position_marker
        )
        return (
            source_provider.get_next_line(),
            new_token,
            3 + len(collected_lines),
            None,
        )

    @staticmethod
    def __is_front_matter_valid(collected_lines):

        ascii_letters_and_digits = f"{string.ascii_letters}{string.digits}_-"

        current_title = None
        current_value = None
        value_map = {}

        for next_line in collected_lines:
            POGGER.debug("Next fm:>$s<", next_line)
            next_index, _ = ParserHelper.extract_whitespace(next_line, 0)
            if next_index >= 4:
                POGGER.debug("Indented line established.")
                if current_title:
                    current_value += "\n" + next_line.strip()
                    POGGER.debug("current_value>$<", current_value)
                else:
                    return "Continuation line encountered before a keyword line."
            else:
                if not next_line.strip():
                    return "Blank line encountered before end of metadata."

                POGGER.debug("Non-indented line established.")
                if current_title:
                    POGGER.debug("Adding '$' as '$'.", current_title, current_value)
                    value_map[current_title] = current_value

                (
                    next_index,
                    current_title,
                ) = ParserHelper.collect_while_one_of_characters(
                    next_line, next_index, ascii_letters_and_digits
                )
                if next_index < len(next_line) and next_line[next_index] == ":":
                    current_value = next_line[next_index + 1 :].strip()
                else:
                    return "Newline did not start with `keyword:`."
        if current_title:
            POGGER.debug("Adding final '$' as '$'.", current_title, current_value)
            value_map[current_title.lower()] = current_value
        if not value_map:
            return "No valid metadata header lines were found."
        return value_map


class FrontMatterMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the front matter data.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        start_boundary_line,
        end_boundary_line,
        collected_lines,
        matter_map,
        position_marker,
    ):
        self.__start_boundary_line = start_boundary_line
        self.__end_boundary_line = end_boundary_line
        self.__collected_lines = collected_lines
        self.__matter_map = matter_map

        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_front_matter,
            "",
            position_marker=position_marker,
            is_extension=True,
        )
        self.__compose_extra_data_field()

    # pylint: enable=too-many-arguments

    @property
    def start_boundary_line(self):
        """
        Returns the boundary line used to start the front matter block.
        """
        return self.__start_boundary_line

    @property
    def end_boundary_line(self):
        """
        Returns the boundary line used to stop the front matter block.
        """
        return self.__end_boundary_line

    @property
    def collected_lines(self):
        """
        Returns the collected lines that comprise the front matter block.
        """
        return self.__collected_lines

    @property
    def matter_map(self):
        """
        Returns the processed lines from the front matter block.
        """
        return self.__matter_map

    def __compose_extra_data_field(self):
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        self._set_extra_data(
            MarkdownToken.extra_data_separator.join(
                [
                    self.__start_boundary_line,
                    self.__end_boundary_line,
                    str(self.__collected_lines),
                    str(self.__matter_map),
                ]
            )
        )

    @classmethod
    def calculate_block_token_height(cls, last_token):
        """
        Calculate the height of the token with the given properties.
        """
        return 2 + len(last_token.collected_lines)

    @classmethod
    def calculate_initial_whitespace(cls):
        """
        Calculate the amount of whitespace for the token.
        """
        return 0, False



"""
https://github.github.com/gfm/#thematic-breaks
"""
import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_thematic_breaks_013():
    """
    Test case 013:  A line consisting of 0-3 spaces of indentation, followed by a sequence of three or more matching -, _, or * characters, each followed optionally by any number of spaces or tabs, forms a thematic break.
    """

    # Arrange
    source_markdown = """***
---
___"""
    expected_tokens = [
        "[tbreak(1,1):*::***]",
        "[tbreak(2,1):-::---]",
        "[tbreak(3,1):_::___]",
    ]
    expected_gfm = """<hr />
<hr />
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_014():
    """
    Test case 014:  (part a) Wrong characters:
    """

    # Arrange
    source_markdown = """+++"""
    expected_tokens = ["[para(1,1):]", "[text(1,1):+++:]", "[end-para:::True]"]
    expected_gfm = """<p>+++</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_015():
    """
    Test case 015:  (part b) Wrong characters:
    """

    # Arrange
    source_markdown = """==="""
    expected_tokens = ["[para(1,1):]", "[text(1,1):===:]", "[end-para:::True]"]
    expected_gfm = """<p>===</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_016():
    """
    Test case 016:  Not enough characters:
    """

    # Arrange
    source_markdown = """--
**
__"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):--\n::\n]",
        "[text(2,1):**:]",
        "[text(2,3):\n::\n]",
        "[text(3,1):__:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>--
**
__</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_017():
    """
    Test case 017:  One to three spaces indent are allowed:
    """

    # Arrange
    source_markdown = """ ***
  ***
   ***"""
    expected_tokens = [
        "[tbreak(1,2):*: :***]",
        "[tbreak(2,3):*:  :***]",
        "[tbreak(3,4):*:   :***]",
    ]
    expected_gfm = """<hr />
<hr />
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_018():
    """
    Test case 018:  (part a) Four spaces is too many:
    """

    # Arrange
    source_markdown = """    ***"""
    expected_tokens = [
        "[icode-block(1,5):    :]",
        "[text(1,5):***:]",
        "[end-icode-block:::True]",
    ]
    expected_gfm = """<pre><code>***
</code></pre>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_019():
    """
    Test case 019:  (part b) Four spaces is too many:
    """

    # Arrange
    source_markdown = """Foo
    ***"""
    expected_tokens = [
        "[para(1,1):\n    ]",
        "[text(1,1):Foo\n::\n]",
        "[text(2,5):***:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo
***</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_020():
    """
    Test case 020:  More than three characters may be used:
    """

    # Arrange
    source_markdown = """_____________________________________"""
    expected_tokens = ["[tbreak(1,1):_::_____________________________________]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_021():
    """
    Test case 021:  (part a) Spaces are allowed between the characters:
    """

    # Arrange
    source_markdown = """ - - -"""
    expected_tokens = ["[tbreak(1,2):-: :- - -]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_022():
    """
    Test case 022:  (part b) Spaces are allowed between the characters:
    """

    # Arrange
    source_markdown = """ **  * ** * ** * **"""
    expected_tokens = ["[tbreak(1,2):*: :**  * ** * ** * **]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_023():
    """
    Test case 023:  (part c) Spaces are allowed between the characters:
    """

    # Arrange
    source_markdown = """-     -      -      -"""
    expected_tokens = ["[tbreak(1,1):-::-     -      -      -]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_024():
    """
    Test case 024:  Spaces are allowed at the end:
    """

    # Arrange
    source_markdown = """- - - -    """
    expected_tokens = ["[tbreak(1,1):-::- - - -    ]"]
    expected_gfm = """<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_025():
    """
    Test case 025:  However, no other characters may occur in the line:
    """

    # Arrange
    source_markdown = """_ _ _ _ a

a------

---a---"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):_:]",
        "[text(1,2): :]",
        "[text(1,3):_:]",
        "[text(1,4): :]",
        "[text(1,5):_:]",
        "[text(1,6): :]",
        "[text(1,7):_:]",
        "[text(1,8): a:]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[text(3,1):a------:]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[text(5,1):---a---:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>_ _ _ _ a</p>
<p>a------</p>
<p>---a---</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_026():
    """
    Test case 026:  It is required that all of the non-whitespace characters be the same. So, this is not a thematic break:
    """

    # Arrange
    source_markdown = """ *-*"""
    expected_tokens = [
        "[para(1,2): ]",
        "[emphasis(1,2):1:*]",
        "[text(1,3):-:]",
        "[end-emphasis(1,4)::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>-</em></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_027():
    """
    Test case 027:  Thematic breaks do not need blank lines before or after:
    """

    # Arrange
    source_markdown = """- foo
***
- bar"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):*::***]",
        "[ulist(3,1):-::2:]",
        "[para(3,3):]",
        "[text(3,3):bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>foo</li>
</ul>
<hr />
<ul>
<li>bar</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_028():
    """
    Test case 028:  Thematic breaks can interrupt a paragraph:
    """

    # Arrange
    source_markdown = """Foo
***
bar"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):Foo:]",
        "[end-para:::False]",
        "[tbreak(2,1):*::***]",
        "[para(3,1):]",
        "[text(3,1):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>Foo</p>
<hr />
<p>bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_029():
    """
    Test case 029:  If a line of dashes that meets the above conditions for being a thematic break could also be interpreted as the underline of a setext heading, the interpretation as a setext heading takes precedence. Thus, for example, this is a setext heading, not a paragraph followed by a thematic break:
    """

    # Arrange
    source_markdown = """Foo
---
bar"""
    expected_tokens = [
        "[setext(2,1):-:3::(1,1)]",
        "[text(1,1):Foo:]",
        "[end-setext::]",
        "[para(3,1):]",
        "[text(3,1):bar:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<h2>Foo</h2>
<p>bar</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_030():
    """
    Test case 030:  When both a thematic break and a list item are possible interpretations of a line, the thematic break takes precedence:
    """

    # Arrange
    source_markdown = """* Foo
* * *
* Bar"""
    expected_tokens = [
        "[ulist(1,1):*::2:]",
        "[para(1,3):]",
        "[text(1,3):Foo:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(2,1):*::* * *]",
        "[ulist(3,1):*::2:]",
        "[para(3,3):]",
        "[text(3,3):Bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>Foo</li>
</ul>
<hr />
<ul>
<li>Bar</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_031():
    """
    Test case 031:  If you want a thematic break in a list item, use a different bullet:
    """

    # Arrange
    source_markdown = """
- Foo
- * * *"""
    expected_tokens = [
        "[BLANK(1,1):]",
        "[ulist(2,1):-::2:]",
        "[para(2,3):]",
        "[text(2,3):Foo:]",
        "[end-para:::True]",
        "[li(3,1):2::]",
        "[tbreak(3,3):*::* * *]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>Foo</li>
<li>
<hr />
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_thematic_breaks_031o():
    """
    Test case 031:  If you want a thematic break in a list item, use a different bullet:
    """

    # Arrange
    source_markdown = """
1) Foo
1) * * *"""
    expected_tokens = [
        "[BLANK(1,1):]",
        "[olist(2,1):):1:3:]",
        "[para(2,4):]",
        "[text(2,4):Foo:]",
        "[end-para:::True]",
        "[li(3,1):3::1]",
        "[tbreak(3,4):*::* * *]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>Foo</li>
<li>
<hr />
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide tests related to the MD022 rule.
"""
import os
from test.markdown_scanner import MarkdownScanner

import pytest

from .utils import write_temporary_configuration

# pylint: disable=too-many-lines


@pytest.mark.rules
def test_md022_bad_proper_line_spacing_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that have proper default
    spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/proper_line_spacing_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_proper_line_spacing_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that have proper default
    spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/proper_line_spacing_setext.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_atx.md:1:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx.md:4:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx.md:4:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_atx_in_same_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_atx_in_same_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_atx_in_same_block_quote.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_same_block_quote.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_same_block_quote.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_atx_in_same_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_atx_in_same_list_item.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_atx_in_same_list_item.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_same_list_item.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_same_list_item.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_atx_in_different_list_items():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing both before and after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_atx_in_different_list_items.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_atx_in_different_list_items.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_different_list_items.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_atx_in_different_list_items.md:4:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_before_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_before_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_before_atx.md:5:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_before_atx_in_same_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_before_atx_in_same_list_item.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_before_atx_in_same_list_item.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_before_atx_in_different_list_items():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_before_atx_in_different_list_items.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_before_atx_in_different_list_items.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_before_atx_in_same_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_before_atx_in_same_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_before_atx_in_same_block_quote.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_before_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_before_setext.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_atx.md:1:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_atx.md:5:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_atx_in_same_list_item():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_atx_in_same_list_item.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_atx_in_same_list_item.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_atx_in_same_list_item.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_atx_in_same_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_atx_in_same_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_atx_in_same_block_quote.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_atx_in_same_block_quote.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_atx_in_different_list_items():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_atx_in_different_list_items.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_atx_in_different_list_items.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_atx_in_different_list_items.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_atx_in_different_block_quotes():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing after the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md028",
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_atx_in_different_block_quotes.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_atx_in_different_block_quotes.md:1:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_atx_in_different_block_quotes.md:5:3: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_atx_with_html_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_html_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_atx_with_html_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_html_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/atx_with_html_and_bad_line_spacing.md:3:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/atx_with_html_and_bad_line_spacing.md:8:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_atx_with_paragraph_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_paragraph_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_atx_with_paragraph_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_paragraph_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/atx_with_paragraph_and_bad_line_spacing.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/atx_with_paragraph_and_bad_line_spacing.md:7:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_atx_with_code_block_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    default spacing before and after the heading, surrounded by code blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md040",
        "scan",
        "test/resources/rules/md022/atx_with_code_block_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_atx_with_code_block_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before and after the heading, surrounded by code blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md040,md031",
        "scan",
        "test/resources/rules/md022/atx_with_code_block_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/atx_with_code_block_and_bad_line_spacing.md:4:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/atx_with_code_block_and_bad_line_spacing.md:9:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_atx_with_thematic_break_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    default spacing before and after the heading, surrounded by thematic breaks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_thematic_break_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_atx_with_thematic_break_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    default spacing before and after the heading, surrounded by thematic breaks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/atx_with_thematic_break_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/atx_with_thematic_break_and_bad_line_spacing.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/atx_with_thematic_break_and_bad_line_spacing.md:7:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do have proper
    default spacing before the heading.

    Note that setext grabs the last paragraph before the marker and puts it as the
    heading.  As such, testing this for one line space before is implied as one line
    space is required to break up the paragraph.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_setext.md:1:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_setext.md:6:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_no_line_spacing_after_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do not have proper
    default spacing before the heading.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/no_line_spacing_after_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/no_line_spacing_after_setext.md:1:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/no_line_spacing_after_setext.md:6:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_setext_with_code_block_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do have proper
    default spacing before and after the heading, surrounded by code blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md040",
        "scan",
        "test/resources/rules/md022/setext_with_code_block_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_setext_with_code_block_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do not have proper
    default spacing before and after the heading, surrounded by code blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md040,md031",
        "scan",
        "test/resources/rules/md022/setext_with_code_block_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/setext_with_code_block_and_bad_line_spacing.md:4:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/setext_with_code_block_and_bad_line_spacing.md:10:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_setext_with_html_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/setext_with_html_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_setext_with_html_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do not have proper
    default spacing before and after the heading, surrounded by html blocks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/setext_with_html_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/setext_with_html_and_bad_line_spacing.md:3:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/setext_with_html_and_bad_line_spacing.md:9:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_setext_with_thematic_break_and_good_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do have proper
    default spacing before and after the heading, surrounded by thematic breaks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/setext_with_thematic_break_and_good_line_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_setext_with_thematic_break_and_bad_line_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has setext headings that do not have proper
    default spacing before and after the heading, surrounded by thematic breaks.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/setext_with_thematic_break_and_bad_line_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/setext_with_thematic_break_and_bad_line_spacing.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/setext_with_thematic_break_and_bad_line_spacing.md:8:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_proper_line_spacing_atx_with_alternate_lines_above():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    spacing before the heading with an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md022": {"lines_above": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/proper_line_spacing_atx.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md022/proper_line_spacing_atx.md:7:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_good_double_line_spacing_above_atx_with_alternate_lines_above():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    spacing before the heading with an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md022": {"lines_above": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "md012",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/double_line_spacing_above_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_bad_proper_line_spacing_atx_with_alternate_lines_below():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do not have proper
    spacing below the heading with an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md022": {"lines_below": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/proper_line_spacing_atx.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md022/proper_line_spacing_atx.md:1:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/proper_line_spacing_atx.md:7:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_good_double_line_spacing_above_atx_with_alternate_lines_below():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    spacing below the heading with an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {"plugins": {"md022": {"lines_below": 2}}}
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "md012",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/double_line_spacing_below_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_good_double_line_spacing_above_and_below_atx_with_alternate_lines_both():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has atx headings that do have proper
    spacing above and below the heading with an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {
        "plugins": {"md022": {"lines_below": 2, "lines_above": 2}}
    }
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "md012",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/double_line_spacing_above_and_below_atx.md",
        ]

        expected_return_code = 0
        expected_output = ""
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_good_alternating_heading_types():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has alternating heading types.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "MD003,md025",
        "scan",
        "test/resources/rules/md022/alternating_heading_types.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_alternating_heading_types_with_alternate_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has alternating heading types with
    an alternate configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {
        "plugins": {"md022": {"lines_below": 2, "lines_above": 2}}
    }
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "MD003,md025",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/alternating_heading_types.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md022/alternating_heading_types.md:1:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:3:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:3:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:6:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:6:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:8:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:8:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Below] (blanks-around-headings,blanks-around-headers)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_bad_alternating_heading_types_with_alternate_spacing_and_bad_config():
    """
    Same as above, but with the lines_below being an illegal value.  Without
    strict mode, this means the value will not be used, and the default of 1
    will be used.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {
        "plugins": {"md022": {"lines_below": -2, "lines_above": 2}}
    }
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--disable-rules",
            "MD003,md025",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/alternating_heading_types.md",
        ]

        expected_return_code = 1
        expected_output = (
            "test/resources/rules/md022/alternating_heading_types.md:3:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:6:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
            + "test/resources/rules/md022/alternating_heading_types.md:8:1: "
            + "MD022: Headings should be surrounded by blank lines. "
            + "[Expected: 2; Actual: 1; Above] (blanks-around-headings,blanks-around-headers)\n"
        )
        expected_error = ""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_bad_alternating_heading_types_with_alternate_spacing_and_bad_config_strict_mode():
    """
    Same as above, but with the lines_below being an illegal value.  With strict mode,
    this means that an error will be displayed.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {
        "plugins": {"md022": {"lines_below": -2, "lines_above": 2}}
    }
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "--strict-config",
            "--disable-rules",
            "MD003",
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md022/alternating_heading_types.md",
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = (
            "BadPluginError encountered while configuring plugins:\n"
            + "The value for property 'plugins.md022.lines_below' is not valid: Value must not be zero or a positive integer.\n"
        )

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


@pytest.mark.rules
def test_md022_good_unordered_list_into_atx_into_paragraph():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has an atx heading after an unsorted
    list and a paragraph following it.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/unordered_list_into_atx_into_paragraph.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_heading_surrounded_by_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has an atx heading after an unsorted
    list and a paragraph following it.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/bad_heading_surrounded_by_block_quote.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/bad_heading_surrounded_by_block_quote.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/bad_heading_surrounded_by_block_quote.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_bad_heading_surrounded_by_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has an atx heading after an unsorted
    list and a paragraph following it.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md032",
        "scan",
        "test/resources/rules/md022/bad_heading_surrounded_by_list.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md022/bad_heading_surrounded_by_list.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Above] (blanks-around-headings,blanks-around-headers)\n"
        + "test/resources/rules/md022/bad_heading_surrounded_by_list.md:2:1: "
        + "MD022: Headings should be surrounded by blank lines. "
        + "[Expected: 1; Actual: 0; Below] (blanks-around-headings,blanks-around-headers)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_heading_in_block_quote():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has an atx heading after an unsorted
    list and a paragraph following it.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/good_heading_in_block_quote.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md022_good_heading_in_list():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/MD022 directory that has an atx heading after an unsorted
    list and a paragraph following it.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md022/good_heading_in_list.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#paragraph
"""
import pytest

from .utils import act_and_assert

# pylint: disable=too-many-lines


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_t_nl_hs():
    """
    Test case:  Ordered list newline text new line setext heading
    was:        test_list_blocks_256hx
    """

    # Arrange
    source_markdown = """1.
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[setext(3,1):-:3::(2,1)]",
        "[text(2,1):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_t_nl_all_hs():
    """
    Test case:  Ordered list newline text new line setext heading
    """

    # Arrange
    source_markdown = """1.
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[BLANK(1,3):]",
        "[setext(3,4):-:3::(2,4)]",
        "[text(2,4):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h2>foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i2_t_nl_hs():
    """
    Test case:  Ordered list newline indent of 2 text new line setext heading
    was:        test_list_blocks_256hxa
    """

    # Arrange
    source_markdown = """1.
  foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[end-olist:::True]",
        "[setext(3,1):-:3:  :(2,3)]",
        "[text(2,3):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li></li>
</ol>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_t_nl_hs():
    """
    Test case:  Ordered list newline indent of 3 text new line setext heading
    was:        test_list_blocks_256hxb
    """

    # Arrange
    source_markdown = """1.
   foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[para(2,4):]",
        "[text(2,4):foo:]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(3,1):-::---]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list newline indent of 3 text new line indent of 3 setext heading
    was:        test_list_blocks_256hxc
    """

    # Arrange
    source_markdown = """1.
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[BLANK(1,3):]",
        "[setext(3,4):-:3::(2,4)]",
        "[text(2,4):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h2>foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_t_nl_hs():
    """
    Test case:  Ordered list text newline text new line setext heading
    was:        test_list_blocks_256ha
    """

    # Arrange
    source_markdown = """1. abc
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::]",
        "[para(1,4):\n]",
        "[text(1,4):abc\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(3,1):-::---]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_t_nl_all_hs():
    """
    Test case:  Ordered list text newline text new line setext heading
    """

    # Arrange
    source_markdown = """1. abc
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[setext(3,4):-:3::(1,4)]",
        "[text(1,4):abc\nfoo::\n]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h2>abc
foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i2_t_nl_hs():
    """
    Test case:  Ordered list text newline indent of 2 text new line setext heading
    was:        test_list_blocks_256haa
    """

    # Arrange
    source_markdown = """1. abc
  foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::  ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(3,1):-::---]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_t_nl_hs():
    """
    Test case:  Ordered list text newline indent of 3 text new line setext heading
    was:        test_list_blocks_256hab
    """

    # Arrange
    source_markdown = """1. abc
   foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):\n]",
        "[text(1,4):abc\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(3,1):-::---]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list text newline indent of 3 text new line indent of 3 setext heading
    was:        test_list_blocks_256hac
    """

    # Arrange
    source_markdown = """1. abc
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[setext(3,4):-:3::(1,4)]",
        "[text(1,4):abc\nfoo::\n]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h2>abc
foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_ol_nl_t_nl_hs():
    """
    Test case:  Ordered list x2 text newline new line setext heading
    """

    # Arrange
    source_markdown = """1. 1.
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   ]",
        "[BLANK(1,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[setext(3,1):-:3::(2,1)]",
        "[text(2,1):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_ol_nl_t_nl_all_hs():
    """
    Test case:  Ordered list x2 text newline new line setext heading
    """

    # Arrange
    source_markdown = """1. 1.
      foo
      ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   :      \n      ]",
        "[BLANK(1,6):]",
        "[setext(3,7):-:3::(2,7)]",
        "[text(2,7):foo:]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>
<h2>foo</h2>
</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_ol_t_nl_t_nl_hs():
    """
    Test case:  Ordered list x2 text newline text new line setext heading
    was:        test_list_blocks_256hb
    """

    # Arrange
    source_markdown = """1. 1. abc
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   :]",
        "[para(1,7):\n]",
        "[text(1,7):abc\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[tbreak(3,1):-::---]",
        "[BLANK(4,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>abc
foo</li>
</ol>
</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_ol_t_nl_t_nl_all_hs():
    """
    Test case:  Ordered list x2 text newline text new line setext heading
    """

    # Arrange
    source_markdown = """1. 1. abc
      foo
      ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[olist(1,4):.:1:6:   :      \n      ]",
        "[setext(3,7):-:3::(1,7)]",
        "[text(1,7):abc\nfoo::\n]",
        "[end-setext::]",
        "[BLANK(4,1):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>
<h2>abc
foo</h2>
</li>
</ol>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_nl_t_nl_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list new line text newline setext heading
    """

    # Arrange
    source_markdown = """1.
   1.
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[setext(4,1):-:3::(3,1)]",
        "[text(3,1):foo:]",
        "[end-setext::]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_nl_t_nl_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list new line text newline setext heading
    """

    # Arrange
    source_markdown = """1. abc
   1.
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n]",
        "[para(1,4):\n\n]",
        "[text(1,4):abc\n1.\nfoo::\n\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.
foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ul_nl_t_nl_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list new line text newline setext heading
    """

    # Arrange
    source_markdown = """- abc
  -
foo
---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[setext(4,1):-:3::(3,1)]",
        "[text(3,1):foo:]",
        "[end-setext::]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ulb_nl_t_nl_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) new line text newline setext heading
    """

    # Arrange
    source_markdown = """- abc
  *
foo
---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n]",
        "[para(1,3):\n\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[text(2,2):\nfoo::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*
foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_t_nl_t_nl_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list text new line text newline setext heading
    """

    # Arrange
    source_markdown = """1.
   1. def
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   :]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def
foo</li>
</ol>
</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_t_nl_t_nl_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text new line text newline setext heading
    was:        test_list_blocks_256hc
    """

    # Arrange
    source_markdown = """1. abc
   1. def
foo
---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   :]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[tbreak(4,1):-::---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def
foo</li>
</ol>
</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_nl_i2_t_nl_i2_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list new line indent of 2 text newline indent of 2 setext heading
    """

    # Arrange
    source_markdown = """1.
   1.
  foo
  ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[setext(4,3):-:3:  :(3,3)]",
        "[text(3,3):foo:]",
        "[end-setext:  :]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
</li>
</ol>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_nl_i2_t_nl_i2_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list new line indent of 2 text newline indent of 2 setext heading
    """

    # Arrange
    source_markdown = """1. abc
   1.
  foo
  ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n  ]",
        "[para(1,4):\n\n]",
        "[text(1,4):abc\n1.\nfoo::\n\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(4,3):-:  :---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
1.
foo</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ul_nl_i1_t_nl_i1_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list new line indent of 1 text newline indent of 1 setext heading
    """

    # Arrange
    source_markdown = """- abc
  -
 foo
 ---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[end-ulist:::True]",
        "[setext(4,2):-:3: :(3,2)]",
        "[text(3,2):foo:]",
        "[end-setext: :]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
</li>
</ul>
<h2>foo</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ulb_nl_i1_t_nl_i1_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) new line indent of 1 text newline indent of 1 setext heading
    """

    # Arrange
    source_markdown = """- abc
  *
 foo
 ---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n ]",
        "[para(1,3):\n\n]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[text(2,2):\nfoo::\n]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[tbreak(4,2):-: :---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ul>
<li>abc
*
foo</li>
</ul>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_t_nl_i2_t_nl_i2_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list text new line indent of 2 text newline indent of 2 setext heading
    """

    # Arrange
    source_markdown = """1.
   1. def
  foo
  ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   :  ]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[tbreak(4,3):-:  :---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def
foo</li>
</ol>
</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_t_nl_i2_t_nl_i2_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text new line indent of 2 text newline indent of 2 setext heading
    was:        test_list_blocks_256hd
    """

    # Arrange
    source_markdown = """1. abc
   1. def
  foo
  ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3:]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   :  ]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[end-olist:::True]",
        "[tbreak(4,3):-:  :---]",
        "[BLANK(5,1):]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def
foo</li>
</ol>
</li>
</ol>
<hr />"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list new line indent of 3 text newline indent of 3 setext heading
    """

    # Arrange
    source_markdown = """1.
   1.
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   ]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   ]",
        "[BLANK(2,6):]",
        "[end-olist:::True]",
        "[setext(4,4):-:3::(3,4)]",
        "[text(3,4):foo:]",
        "[end-setext::]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li></li>
</ol>
<h2>foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list new line indent of 3 text newline indent of 3 setext heading
    """

    # Arrange
    source_markdown = """1. abc
   1.
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   \n   \n   ]",
        "[setext(4,4):-:3::(1,4)]",
        "[text(1,4):abc\n1.\nfoo::\n\n]",
        "[end-setext::]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<h2>abc
1.
foo</h2>
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ul_nl_i2_t_nl_i2_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list new line indent of 2 text newline indent of 2 setext heading
    """

    # Arrange
    source_markdown = """- abc
  -
  foo
  ---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[setext(2,3):-:1::(1,3)]",
        "[text(1,3):abc:]",
        "[end-setext::]",
        "[setext(4,3):-:3::(3,3)]",
        "[text(3,3):foo:]",
        "[end-setext::]",
        "[BLANK(5,1):]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc</h2>
<h2>foo</h2>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ul_t_nl_i2_ulb_nl_i2_t_nl_i2_hs():
    """
    Test case:  Unordered list text newline indent of 2 unordered list (b) new line indent of 2 text newline indent of 2 setext heading
    """

    # Arrange
    source_markdown = """- abc
  *
  foo
  ---
"""
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  \n  ]",
        "[setext(4,3):-:3::(1,3)]",
        "[text(1,3):abc\n::\n]",
        "[text(2,1):*:]",
        "[text(2,2):\nfoo::\n]",
        "[end-setext::]",
        "[BLANK(5,1):]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li>
<h2>abc
*
foo</h2>
</li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_nl_i3_ol_t_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list newline indent of 3 ordered list text new line indent of 3 text newline indent of 3 setext heading
    """

    # Arrange
    source_markdown = """1.
   1. def
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[BLANK(1,3):]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(4,4):-::---]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>
<ol>
<li>def
foo</li>
</ol>
<hr />
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_paragraph_series_m_hs_ol_t_nl_i3_ol_t_nl_i3_t_nl_i3_hs():
    """
    Test case:  Ordered list text newline indent of 3 ordered list text new line indent of 3 text newline indent of 3 setext heading
    was:        test_list_blocks_256he
    """

    # Arrange
    source_markdown = """1. abc
   1. def
   foo
   ---
"""
    expected_tokens = [
        "[olist(1,1):.:1:3::   ]",
        "[para(1,4):]",
        "[text(1,4):abc:]",
        "[end-para:::True]",
        "[olist(2,4):.:1:6:   :   ]",
        "[para(2,7):\n]",
        "[text(2,7):def\nfoo::\n]",
        "[end-para:::True]",
        "[end-olist:::True]",
        "[tbreak(4,4):-::---]",
        "[BLANK(5,1):]",
        "[end-olist:::True]",
    ]
    expected_gfm = """<ol>
<li>abc
<ol>
<li>def
foo</li>
</ol>
<hr />
</li>
</ol>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)
